#include <math.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <signal.h>
#include <assert.h>
#ifndef OLD_SUNOS
#include <sys/audioio.h>
#endif
#include "music.h"

#define TWOPI (2*M_PI)
#define SAMPLERATE 8000			/* Sun's sampling rate */
#define TSIZE 65536			/* sinewave table size */
#define BSIZE 8192			/* # of sound samples per second */
#define MAX_NOTES 3			/* max # of notes to accept at once */

static double sinewave[TSIZE];          /* nice big sinewave */
static double ibuf[BSIZE];              /* raw linear amplitude */
static unsigned char obuf[BSIZE];       /* ulaw samples - check range */
extern char linear_to_ulaw[];           /* Mouse's conversion table */


typedef struct _Node {
  unsigned int freq;
  struct _Node* next;
} Node;

static Node* g_List = NULL;
int g_bInitialized = 0;
pid_t g_pid = 0;


void GenerateSound(Node* List) {

  Node* n;
  int fd;
  int gap;
  int i;
  int dc = 16384;                 /* DC component of amplitude */
  int j;                          /* ulaw converted sample */
  int count = 0;
  
  assert(List);
  
  /* zero the input buffer */
  for (i = 0; i < BSIZE; i++)
    ibuf[i] = 0;

  n = List;
  printf("List of freq: {\n");
  while (n->next) {
    printf("%d\n",n->next->freq);
    gap = (TSIZE * n->next->freq) / SAMPLERATE;

    /* sum contributions from each note's samples */
    for (i = 0; i < BSIZE; i++)
      ibuf[i] += sinewave[(i*gap)%TSIZE];
    
    n = n->next;
    count++;
  }
  printf("}\n\n");
  
  /* at this point, input buffer contains a sum
   * of sinewave samples (ranging from -1 to 1);
   * for volume control, must normalize
   */
  for (i = 0; i < BSIZE; i++) {
#ifdef TRUE_VOLUME
    j = dc  * (1 + ibuf[i] / NOTES);
#else
    j = dc  * (1 + ibuf[i] / count);
#endif
    assert (j >= 0);
    assert (j <= 65535);
    obuf[i] = linear_to_ulaw[j];
  }
                   
                        
  fd = open("/dev/audio",O_WRONLY);
  if (-1 == fd) {
    perror("open for /dev/audio");
    exit(-1);
  }
  
  while (1) {
    write(fd, obuf, BSIZE);
#ifndef OLD_SUNOS
    ioctl(fd,AUDIO_DRAIN); /* wait for audio to be sent... */
#else
    sleep(1);
#endif
  }
}

  
int RestartSoundGen() {

  int rc;
  
  if (g_pid != 0) {
    rc = kill(g_pid,SIGKILL);
    if (-1 == rc)
      return E_KILL;
  }

  g_pid = 0;

  if (g_List->next) { /* if we have sounds ... */
    
    rc = fork();
    if (-1 == rc)
      return E_FORK;

    g_pid = rc;
    
    if (g_pid == 0) {
      GenerateSound(g_List);
      exit(0); /* GenerateSound runs forever, but just to be safe */
    }
  }
  
  return OK;
  
}

Node* CreateNode(unsigned int freq,Node* next) {
  Node* result;

  result = malloc(sizeof(Node));
  if (result) {
    result->freq = freq;
    result->next = next;
  }

  return result;
    
}

Node* FindPrev(Node* List,unsigned int freq) {

  assert(List);

  while (List->next != NULL && List->next->freq != freq)
    List = List->next;

  return List;
}


int InitMusic() {
  int i;

  if (g_bInitialized)
    return E_ALREADY_INITIALIZED;
  
  g_List = CreateNode(0,NULL);
  if (NULL == g_List)
    return E_OUT_OF_MEM;

  /* construct a nice big sine wave  */
  for (i=0;i<TSIZE;i++)
    sinewave[i] = sin((i*TWOPI)/TSIZE);

  g_bInitialized = 1;
  
  return OK;
}

int AddRemoveFreq(unsigned int* Add,unsigned int AddSize,
                  unsigned int* Remove,unsigned int RemoveSize,
                  unsigned int* RemoveResults) {

  unsigned int i;
  Node* n;
  Node* tmp;
  int rc = OK;
  int restartrc = OK;

  if (!g_bInitialized)
    return E_NOT_INITIALIZED;

  if (NULL == Add && AddSize > 0)
    return E_INVALID_ARG;

  if (NULL == Remove && RemoveSize > 0)
    return E_INVALID_ARG;
  
  assert(g_List);

  for (i = 0; i < AddSize; i++) {
    n = CreateNode(Add[i],g_List->next);
    
    if (NULL == n)
      return E_OUT_OF_MEM;
    g_List->next = n;
  }

  for (i = 0; i < RemoveSize; i++) {
    
    n = FindPrev(g_List,Remove[i]);

    if (NULL == n->next) {
      rc = E_FREQ_NOT_FOUND;
      if (RemoveResults)
        RemoveResults[i] = E_FREQ_NOT_FOUND;
    }
    else {
      if (RemoveResults)
        RemoveResults[i] = OK;
      tmp = n->next;
      n->next = n->next->next;
      free(tmp);
    }
  }

  restartrc = RestartSoundGen();

  return (OK == restartrc) ? rc : restartrc;
}

/*

int StartFreq(unsigned int freq) {
  Node* n;

  if (!g_bInitialized)
    return E_NOT_INITIALIZED;
  
  assert(g_List);
  
  n = CreateNode(freq,g_List->next);
  if (NULL == n)
    return E_OUT_OF_MEM;
  g_List->next = n;
  
  return RestartSoundGen();
}

int StopFreq(unsigned int freq) {
  Node* n;
  Node* tmp;
  
  if (!g_bInitialized)
    return E_NOT_INITIALIZED;

  assert(g_List);
  
  n = FindPrev(g_List,freq);

  if (NULL == n->next)
    return E_FREQ_NOT_FOUND;

  tmp = n->next;
  n->next = n->next->next;
  free(tmp);

  return RestartSoundGen();
}
*/
int CleanMusic() {
  Node* tmp;
  int rc;
  
  if (!g_bInitialized)
    return E_NOT_INITIALIZED;

  assert(g_List);
  
  while (g_List) {
    tmp = g_List;
    g_List = g_List->next;
    free(tmp);
  }

  if (g_pid != 0) {
    rc = kill(g_pid,SIGTERM);
    if (-1 == rc)
      return E_KILL;
  }
  
  g_bInitialized = 0;
  
  return OK;
  
}
